//============================================================
void createCachedField(int maxR){
  int		x, y, dx, dy;
  int		size = (maxR*2)+1;
  float  	R2 = maxR*maxR;
  float  	R4 = R2*R2;
  float  	R6 = R4*R2;
  float  	fn = 0 - (4.00/9.0);
  float  	sn =     (17.0/9.0);
  float  	tn = 0 - (22.0/9.0);
  float  	r2, r4, r6;
  float  	val;

  cachedField	= new float[size][size];
  for (y=0; y<size; y++){
    for (x=0; x<size; x++){
      dx = maxR - x;
      dy = maxR - y;
      r2 = (dx*dx + dy*dy);
      r4 = r2*r2;
      r6 = r4*r2;

      if (r2 <= R2){ 
        val = (fn*r6/R6)+(sn*r4/R4)+(tn*r2/R2)+1.0;
      } 
      else { 
        val = 0; 
      }
      cachedField[y][x] = (val * maxVal);
    }
  }
}


//============================================================
byte evaluateSquare(int col, int row){

  byte b = 0;
  int lef = col*squareSize;
  int top = row*squareSize;
  int rit = lef + squareSize;
  int bot = top + squareSize;

  int frb = getFieldValueAtPixel(rit,bot); 
  int frt = getFieldValueAtPixel(rit,top); 
  int flt = getFieldValueAtPixel(lef,top);
  int flb = getFieldValueAtPixel(lef,bot);

  if ((frb < threshHold)){ 
    b|=(1<<0); 
  }
  if ((frt < threshHold)){ 
    b|=(1<<1); 
  }
  if ((flt < threshHold)){ 
    b|=(1<<2); 
  }
  if ((flb < threshHold)){ 
    b|=(1<<3); 
  }

  return b;
}



//============================================================
Point findFirstSquareContainingEdge (int col, int row){
  // given the row and column of a square containing a source,
  // find the row and column of the first-discovered square 
  // near the source which contains the implicit edge
  int i, j;
  int s=0;
  int x=0; 
  int y=0;
  byte b;
  boolean foundSquare = false;

  if (nLissaPaths > 0){ // while loop won't end if there are no sources
    while (foundSquare == false){

      for (i= (col-s); i<=(col+s); i++){ // do horizontal periphery
        j = (row-s); 
        b = evaluateSquare(i, j);
        if (b>0) { 
          foundSquare = true; 
          x=i; 
          y=j;
          break; 
        }

        j = (row+s); 
        b = evaluateSquare(i, j);
        if (b>0) { 
          foundSquare = true; 
          x=i; 
          y=j;
          break; 
        }
      }

      for (j= (row-s+1); j<=(row+s-1); j++){ // do vertical periphery
        i = (col-s); 
        b = evaluateSquare(i, j);
        if (b>0) { 
          foundSquare = true; 
          x=i; 
          y=j; 
          break; 
        }

        i = (col+s); 
        b = evaluateSquare(i, j);
        if (b>0) { 
          foundSquare = true; 
          x=i; 
          y=j; 
          break; 
        }
      }
      s++;
    }
  }
  if (x==0 && y==0){
    // println("bad search");  
  }
  squareReturnPoint.x = x;
  squareReturnPoint.y = y;
  return squareReturnPoint;
}


//============================================================
int getFieldValueAtPixel(int x, int y){
  float val = 0;
  int dx, dy, i, j;
  FPoint fpoint;
  for (int p=0; p<nLissaPaths; p++){
    fpoint = lissaPathArray[p].getCurrentPoint();
    dx = (int)round(x - fpoint.x);
    dy = (int)round(y - fpoint.y);
    if ((Math.abs(dx) <= R) && (Math.abs(dy) <= R)){
      i = R + dx; 
      j = R + dy;
      
      float s = 1.0;
      if (p==21){ //hack for mouseBlob
        s = 0.95;//1.05;
      }
      
      val += s * cachedField[j][i];
      
      
    } 
    else { 
      val += 0;
    }
  }
  return (int) val;
}

//============================================================
Point converge (int col, int row, byte edge){
  // edge must be a byte valued at either 3, 6, 9 or 12
  // and indicates which edge is to be searched for the edge boundary
  Point root;
  int x1 = 0; 
  int x2 = 0;
  int y1 = 0; 
  int y2 = 0;
  float px = 0; 
  float py = 0;
  float lox = 0; 
  float loy = 0;
  float hix = 0; 
  float hiy = 0;

  int lef = col*squareSize;
  int top = row*squareSize;
  int rit = lef + squareSize;
  int bot = top + squareSize;

  switch (edge){
  case 3:	// right edge
    x1 = rit; 
    x2 = rit;
    y1 = bot; 
    y2 = top;
    break;
  case 6:	 // top edge
    x1 = lef; 
    x2 = rit;
    y1 = top; 
    y2 = top;
    break;
  case 9:	 // bottom edge
    x1 = rit; 
    x2 = lef;
    y1 = bot; 
    y2 = bot;
    break;
  case 12: // left edge
    x1 = lef; 
    x2 = lef;
    y1 = bot; 
    y2 = top;
    break;
  }

  if (getFieldValueAtPixel(x1, y1) <= getFieldValueAtPixel(x2, y2)){
    lox = x1; 
    loy = y1; 
    hix = x2; 
    hiy = y2; 
  }
  else { 
    lox = x2; 
    loy = y2; 
    hix = x1; 
    hiy = y1; 
  }

  for (int i=0; i<CONVERGE_TESTS; i++){
    px = (lox + hix)/2.0;
    py = (loy + hiy)/2.0;
    if (getFieldValueAtPixel((int)round(px), (int)round(py)) >= threshHold) { 
      hix = px; 
      hiy = py;
    }
    else {
      lox = px; 
      loy = py;
    }
  }
  rootPoint.x = (int)round(px);
  rootPoint.y = (int)round(py);
  return rootPoint;
}



//============================================================
Point getSquareContainingPixel (int x, int y) {
  // determines the row and column of the square containing the pixel
  //Point p = new Point (x/squareSize, y/squareSize);
  squareContainingPixelPoint.x = x/squareSize;
  squareContainingPixelPoint.y = y/squareSize;
  return squareContainingPixelPoint; //p;
}


//============================================================
Polygon computeBoundary(Point sourceSquare){
  // given a Point containing the row and column
  // of a square containing a source, search outward from the
  // source until we find a square which contains the edge of
  // the implicit curve; then trace along that edge, in search of
  // the other squares which contain the edge; for each such
  // square, represent its row and column in a Point, and store this
  // Point in a polygon
  // Polygon blob;
  Point startSquare;
  Point edgeSquare;
  Point edgePoint;
  boolean incomplete;
  int col, row;
  int prevRow, prevCol;
  int prevPrevRow, prevPrevCol;
  byte edge = 3;
  byte b;

  // blob = new Polygon();
  returnBlobs[CURB].reset();
  startSquare = findFirstSquareContainingEdge(sourceSquare.x, sourceSquare.y);
  prevCol = col = startSquare.x;
  prevRow = row = startSquare.y;

  incomplete = true;
  int nsegs = 1;

  do{
    prevPrevRow = prevRow;
    prevPrevCol = prevCol;
    prevRow = row; 
    prevCol = col;
    b = evaluateSquare(col, row);

    switch(b){

    case 10: 
      boolean comingFromLeft = (prevCol > prevPrevCol); 
      boolean comingFromRight = !comingFromLeft;
      if (comingFromLeft){ // down if coming from left
        row++;
        edge = 9;
      } 
      else if (comingFromRight){ // up if coming from right
        row--;	
        edge = 6;	
      }
      break; 	// UP/DOWN ITHSMUS;
    case 5:  
      boolean comingFromAbove = (prevRow > prevPrevRow);
      boolean comingFromBelow = !comingFromAbove;
      if (comingFromAbove){ // left if coming from above
        col--;	
        edge = 12;
      } 
      else if (comingFromBelow){  // right if coming from below
        col++;
        edge = 3;
      }
      break; 	// LEFT/RIGHT ISTHMUS;

    case 1: 
    case 9:  
    case 13:			
      col++;	
      edge = 3;	
      break;	// to RIGHT  
    case 2: 
    case 3:  
    case 11: 	
      row--;	
      edge = 6;	
      break; 	// to UP;
    case 4:  
    case 6:  
    case 7:	
      col--;	
      edge = 12;	
      break; 	// to LEFT;
    case 8: 
    case 12: 
    case 14:			
      row++;	
      edge = 9;	
      break;	// to DOWN;	
    case 0: 
    case 15: 
      break;
    }

    edgePoint = converge(prevCol, prevRow, edge);
    // edgeSquare = new Point(col, row);
    edgeSquarePoint.x = col;
    edgeSquarePoint.y = row;


    if (nsegs > MAX_BOUNDARY_LENGTH) { 
      incomplete = false; 
      println("failure " + millis());
      // can't find the original startPoint.  abort the search with a hack.
    } 
    else {
      if (!(/*edgeSquare*/ edgeSquarePoint.equals(startSquare))){
        returnBlobs[CURB].addPoint(edgePoint.x, edgePoint.y);
        nsegs++;
      } 
      else { 
        incomplete = false; 
        // edge has looped around. can fail if startPoint was an isthmus...
      } 
    }

  } 
  while (incomplete);
  return returnBlobs[CURB]; 
}



//------------------------------------------------------------
float getPerimeter(Polygon path){
  float pathLength = 0.0F;
  int nPathPoints = path.npoints;
  for (int i = 0; i < nPathPoints; i++){
    float lowerx = path.xpoints[i];
      float lowery = path.ypoints[i];
      float upperx = path.xpoints[(i+1)%nPathPoints];
      float uppery = path.ypoints[(i+1)%nPathPoints];
      float Dx = upperx - lowerx;
      float Dy = uppery - lowery;
    
    float segLength = (float)Math.sqrt(Dx * Dx + Dy * Dy);
    pathLength += segLength;
  }
  return pathLength;
}

//------------------------------------------------------------
int resamplePolygon (Polygon path, Vec2f resampledPath[], int nResamples) {
  
  int nPathPoints = path.npoints;
  float totalPathLength = getPerimeter(path);
  
  int nResampledPoints = (int)min(nResamples, totalPathLength/16.0);

  float RSL = totalPathLength / (float)nResampledPoints;
  float prevRemainder = RSL;
  int p = 0;
  if(nPathPoints <= 1) {
    for(p = 0; p < nResampledPoints; p++){

      float lowerx = path.xpoints[0];
      float lowery = path.ypoints[0];
      
      float px = lowerx + (float)p * 0.0001F;
      float py = lowery + (float)p * 0.0001F;
      resampledPath[p].set(px, py);
    }
  }
  else {
    for (int i = 0; i < nPathPoints; i++){
      
      float lowerx = path.xpoints[i];
      float lowery = path.ypoints[i];
      float upperx = path.xpoints[(i+1)%nPathPoints];
      float uppery = path.ypoints[(i+1)%nPathPoints];
      float Dx = upperx - lowerx;
      float Dy = uppery - lowery;
      
      float segLength = (float)Math.sqrt(Dx * Dx + Dy * Dy);
      float ASL = segLength;
      float dx = Dx / segLength;
      float dy = Dy / segLength;
      float RSLdx = dx * RSL;
      float RSLdy = dy * RSL;
      float neededSpace = RSL - prevRemainder;
      if(ASL >= neededSpace){
        float remainder = ASL;
        float px = lowerx + neededSpace * dx;
        float py = lowery + neededSpace * dy;
        if(p < nResampledPoints) {
          resampledPath[p].set(px, py);
          remainder -= neededSpace;
          p++;
        }
        int nPtsToDo = (int)(remainder / RSL);
        for(int d = 0; d < nPtsToDo; d++) {
          px += RSLdx;
          py += RSLdy;
          if(p < nResampledPoints) {
            resampledPath[p].set(px, py);
            remainder -= RSL;
            p++;
          }
        }

        prevRemainder = remainder;
      }
      else{
        prevRemainder += ASL;
      }
    }
  }
  return nResampledPoints;
}


class FPoint {
  FPoint (){
    x = 0;
    y = 0; 
  }
  FPoint (float xf, float yf){
    x = xf;
    y = yf; 
  }
  float x;
  float y; 
}



//=========================================================================
class Vec2f {
  void set(float X, float Y){
    x = X;
    y = Y;
  }

  Vec2f(float X, float Y){
    set(X, Y);
  }

  Vec2f() {
    set(0.0F, 0.0F);
  }

  float x;
  float y;
}







     /*
        int np = resamplePolygon (blob, resampledPolygon, N_RESAMPLES);
        fill(255);
        smooth();
        stroke(0,0,0);
        strokeWeight(0.5);
        for (int i=0; i<np; i++){
            float px = resampledPolygon[i].x;
            float py = resampledPolygon[i].y;
            ellipse(px, py, 8,8); 
        }
        */
        
       
        
        /*
        strokeWeight(7);
        noFill();
        stroke(0,0,0);
        smooth();
        beginShape(POINTS);
        for (int i=0; i<np; i++){
          vertex(xp[i], yp[i]); 
        }
        endShape(CLOSE);
        */




